from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Pattern as Regex,
    Tuple,
    Union,
    overload,
)
from warnings import warn

import httpx

from .models import CallList, MockResponse, Route, SideEffectError
from .patterns import Pattern, merge_patterns, parse_url_patterns
from .types import (
    ContentDataTypes,
    DefaultType,
    HeaderTypes,
    JSONTypes,
    QueryParamTypes,
    URLPatternTypes,
)


class Router:
    def __init__(
        self,
        *,
        assert_all_called: bool = True,
        assert_all_mocked: bool = True,
        base_url: Optional[str] = None,
    ) -> None:
        self._assert_all_called = assert_all_called
        self._assert_all_mocked = assert_all_mocked
        self._bases = parse_url_patterns(base_url, exact=False)

        self.routes: Dict[Union[str, int], Route] = {}
        self.calls = CallList()
        self.snapshot()

    def clear(self) -> None:
        """
        Clears all routes. May be rolled back to snapshot state.
        """
        self.routes.clear()

    def snapshot(self) -> None:
        """
        Snapshots current routes and calls state.
        """
        # Snapshot current routes and calls
        self._routes = dict(self.routes)
        self._calls = CallList(self.calls)

        # Snapshot each route state
        for route in self._routes.values():
            route.snapshot()

    def rollback(self, reset: bool = True) -> None:
        """
        Rollbacks routes, and optionally calls, to snapshot state.
        """
        # Revert added routes to snapshot
        self.routes.clear()
        self.routes.update(self._routes)

        # Revert each route state to snapshot
        for route in self.routes.values():
            route.rollback(reset=False)

        # Reset call stats to snapshot
        if reset:
            self.reset()

    def reset(self) -> None:
        """
        Resets call stats to snapshot state.
        """
        self.calls[:] = self._calls
        for route in self.routes.values():
            route.reset()

    def assert_all_called(self) -> None:
        assert all(
            (route.called for route in self.routes.values())
        ), "RESPX: some mocked requests were not called!"

    @property
    def stats(self):
        warn(
            ".stats property is deprecated. Please, use .calls",
            category=DeprecationWarning,
        )
        return self.calls

    @property
    def aliases(self):
        warn(
            ".aliases property is deprecated. Please, use .routes",
            category=DeprecationWarning,
        )
        return self.routes

    def __getitem__(self, name: str) -> Optional[Route]:
        return self.routes.get(name)

    @overload
    def pop(self, name: str) -> Route:
        ...  # pragma: nocover

    @overload
    def pop(self, name: str, default: DefaultType) -> Union[Route, DefaultType]:
        ...  # pragma: nocover

    def pop(self, name, default=...):
        """
        Removes a route by name and returns it.

        Raises KeyError when `default` not provided and name is not found.
        """
        try:
            return self.routes.pop(name)
        except KeyError as ex:
            if default is ...:
                raise ex
            return default

    def route(
        self, *patterns: Pattern, name: Optional[str] = None, **lookups: Any
    ) -> Route:
        route = Route(*patterns, **lookups)
        self.add(route, name=name)
        return route

    def add(
        self,
        route: Optional[Union[str, Route, Callable]] = None,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        method: Optional[str] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        """
        Adds a route with given mocked response details.
        """
        if callable(route) and not isinstance(route, Route):
            route = Route(**lookups).mock(side_effect=route)

        elif isinstance(route, str):
            warn(
                "Passing method as string to respx.add is deprecated. "
                "Please, use respx.route(method=...) or respx.get(...)",
                category=DeprecationWarning,
            )
            method = route
            route = None

        if route is None:
            url__lookup = "url__regex" if isinstance(url, Regex) else "url"
            lookups.update({"method": method, url__lookup: url, "params": params})
            route = Route(**lookups)

        response = None
        if (
            status_code is not None
            or headers is not None
            or content_type is not None
            or content is not None
            or text is not None
            or html is not None
            or json is not None
            or pass_through is True
        ):
            if route.side_effect:
                raise NotImplementedError(
                    "Mixing callback with response details is no longer supported"
                )

            warn(
                "Response kwargs among request pattern kwargs is deprecated. "
                "Please, use .respond(...) or % operator.",
                category=DeprecationWarning,
            )

            response = MockResponse(
                status_code,
                headers=headers,
                content_type=content_type,
                content=content,
                text=text,
                html=html,
                json=json,
            )

        if response:
            route.return_value = response

        if pass_through:
            route.pass_through()

        if alias:
            warn(
                'Route alias kwarg is deprecated. Please use name="".',
                category=DeprecationWarning,
            )
            name = alias
        if name:
            route.name = name

        # Merge bases
        route.pattern = merge_patterns(route.pattern, **self._bases)

        route_key = route.name or hash(route)
        if route_key in self.routes:
            # Identical route already exists, swap with new one
            existing_route = self.routes[route_key]
            existing_route.return_value = route.return_value
            existing_route.side_effect = route.side_effect
            existing_route._pass_through = route._pass_through
            route = existing_route
        else:
            self.routes[route_key] = route

        return route

    def get(
        self,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        return self.add(
            method="GET",
            url=url,
            params=params,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
            content=content,
            text=text,
            html=html,
            json=json,
            pass_through=pass_through,
            alias=alias,
            name=name,
            **lookups,
        )

    def post(
        self,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        return self.add(
            method="POST",
            url=url,
            params=params,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
            content=content,
            text=text,
            html=html,
            json=json,
            pass_through=pass_through,
            alias=alias,
            name=name,
            **lookups,
        )

    def put(
        self,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        return self.add(
            method="PUT",
            url=url,
            params=params,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
            content=content,
            text=text,
            html=html,
            json=json,
            pass_through=pass_through,
            alias=alias,
            name=name,
            **lookups,
        )

    def patch(
        self,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        return self.add(
            method="PATCH",
            url=url,
            params=params,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
            content=content,
            text=text,
            html=html,
            json=json,
            pass_through=pass_through,
            alias=alias,
            name=name,
            **lookups,
        )

    def delete(
        self,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        return self.add(
            method="DELETE",
            url=url,
            params=params,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
            content=content,
            text=text,
            html=html,
            json=json,
            pass_through=pass_through,
            alias=alias,
            name=name,
            **lookups,
        )

    def head(
        self,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        return self.add(
            method="HEAD",
            url=url,
            params=params,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
            content=content,
            text=text,
            html=html,
            json=json,
            pass_through=pass_through,
            alias=alias,
            name=name,
            **lookups,
        )

    def options(
        self,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        name: Optional[str] = None,
        **lookups: Any,
    ) -> Route:
        return self.add(
            method="OPTIONS",
            url=url,
            params=params,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
            content=content,
            text=text,
            html=html,
            json=json,
            pass_through=pass_through,
            alias=alias,
            name=name,
            **lookups,
        )

    def record(
        self,
        request: httpx.Request,
        *,
        response: Optional[httpx.Response] = None,
        route: Optional[Route] = None,
    ) -> None:
        call = self.calls.record(request, response)
        if route:
            route.calls.append(call)

    def match(
        self,
        request: httpx.Request,
    ) -> Tuple[Optional[Route], Optional[Union[httpx.Request, httpx.Response]]]:
        route: Optional[Route] = None
        response: Optional[Union[httpx.Request, httpx.Response]] = None

        for prospect in self.routes.values():
            response = prospect.match(request)
            if response:
                route = prospect
                break

        return route, response

    def resolve(self, request: httpx.Request) -> httpx.Response:
        route: Optional[Route] = None
        response: Optional[httpx.Response] = None

        try:
            route, mock = self.match(request)

            if route is None:
                # Assert we always get a route match, if check is enabled
                assert not self._assert_all_mocked, f"RESPX: {request!r} not mocked!"

                # Auto mock a successful empty response
                response = httpx.Response(200)

            elif mock == request:
                # Pass-through request
                response = None

            else:
                # Mocked response
                assert isinstance(mock, httpx.Response)
                response = mock

        except SideEffectError as error:
            self.record(request, response=None, route=error.route)
            raise error.origin
        else:
            self.record(request, response=response, route=route)

        return response



from typing import (
    Any,
    AsyncIterable,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import httpx
from httpcore import AsyncByteStream, SyncByteStream

URL = Tuple[
    bytes,  # scheme
    bytes,  # host
    Optional[int],  # port
    bytes,  # path
]
Headers = List[Tuple[bytes, bytes]]
ByteStream = Union[Iterable[bytes], AsyncIterable[bytes]]
Request = Tuple[
    bytes,  # http method
    URL,
    Headers,
    ByteStream,  # body
]
SyncResponse = Tuple[
    int,  # status code
    Headers,
    SyncByteStream,  # body
    dict,  # ext
]
AsyncResponse = Tuple[
    int,  # status code
    Headers,
    AsyncByteStream,  # body
    dict,  # ext
]
Response = Tuple[
    int,  # status code
    Headers,
    ByteStream,  # body
    dict,  # ext
]

HeaderTypes = Union[
    httpx.Headers,
    Dict[str, str],
    Dict[bytes, bytes],
    Sequence[Tuple[str, str]],
    Sequence[Tuple[bytes, bytes]],
]
CookieTypes = Union[Dict[str, str], Sequence[Tuple[str, str]]]

DefaultType = TypeVar("DefaultType", bound=Any)

Kwargs = Dict[str, Any]
URLPatternTypes = Union[str, Pattern[str], URL, httpx.URL]
JSONTypes = Union[str, List, Dict]
ContentDataTypes = Union[bytes, str, JSONTypes, Callable, Exception]
QueryParamTypes = Union[bytes, str, List[Tuple[str, Any]], Dict[str, Any]]
RequestTypes = Union[Request, httpx.Request]
SideEffectListTypes = Union[httpx.Response, Exception, Type[Exception]]
SideEffectTypes = Union[
    Callable,
    Exception,
    Type[Exception],
    Sequence[SideEffectListTypes],
    Iterator[SideEffectListTypes],
]



import inspect
from functools import partial, wraps
from typing import Any, Callable, Dict, List, Optional, Union
from unittest import mock

from .transports import BaseMockTransport

__all__ = ["MockTransport"]


class MockTransport(BaseMockTransport):
    _local = False
    _patches: List[mock._patch] = []
    transports: List["MockTransport"] = []
    targets = [
        "httpcore._sync.connection.SyncHTTPConnection",
        "httpcore._sync.connection_pool.SyncConnectionPool",
        "httpcore._sync.http_proxy.SyncHTTPProxy",
        "httpcore._async.connection.AsyncHTTPConnection",
        "httpcore._async.connection_pool.AsyncConnectionPool",
        "httpcore._async.http_proxy.AsyncHTTPProxy",
        "httpx._transports.asgi.ASGITransport",
        "httpx._transports.wsgi.WSGITransport",
    ]

    def __call__(
        self,
        func: Optional[Callable] = None,
        assert_all_called: Optional[bool] = None,
        assert_all_mocked: Optional[bool] = None,
        base_url: Optional[str] = None,
    ) -> Union["MockTransport", Callable]:
        """
        Decorator or Context Manager.

        Use decorator/manager with parentheses for local state, or without parentheses
        for global state, i.e. shared patterns added outside of scope.
        """
        if func is None:
            # Parantheses used, branch out to new nested instance.
            # - Only stage when using local ctx `with respx.mock(...) as respx_mock:`
            # - First stage when using local decorator `@respx.mock(...)`
            #   FYI, global ctx `with respx.mock:` hits __enter__ directly
            settings: Dict[str, Any] = {
                "base_url": base_url,
            }
            if assert_all_called is not None:
                settings["assert_all_called"] = assert_all_called
            if assert_all_mocked is not None:
                settings["assert_all_mocked"] = assert_all_mocked
            respx_mock = self.__class__(**settings)
            respx_mock._local = True
            return respx_mock

        # Async Decorator
        @wraps(func)
        async def async_decorator(*args, **kwargs):
            assert func is not None
            if self._local:
                kwargs["respx_mock"] = self
            async with self:
                return await func(*args, **kwargs)

        # Sync Decorator
        @wraps(func)
        def sync_decorator(*args, **kwargs):
            assert func is not None
            if self._local:
                kwargs["respx_mock"] = self
            with self:
                return func(*args, **kwargs)

        # Dispatch async/sync decorator, depening on decorated function.
        # - Only stage when using global decorator `@respx.mock`
        # - Second stage when using local decorator `@respx.mock(...)`
        return async_decorator if inspect.iscoroutinefunction(func) else sync_decorator

    def __enter__(self) -> "MockTransport":
        self.start()
        return self

    def __exit__(self, exception_type: Optional[Exception], *args: Any) -> None:
        self.stop(quiet=bool(exception_type is not None))

    async def __aenter__(self) -> "MockTransport":
        return self.__enter__()

    async def __aexit__(self, *args: Any) -> None:
        self.__exit__(*args)

    def start(self) -> None:
        """
        Register transport and start patching.
        """
        # Idempotent check, i.e. already started
        if self not in self.transports:
            self.snapshot()
            self.transports.append(self)

        self._patch()

    def stop(self, clear: bool = True, reset: bool = True, quiet: bool = False) -> None:
        """
        Unregister transport and stop patching, when no registered transports left.
        """
        started = bool(self in self.transports)

        try:
            if started and not quiet and self._assert_all_called:
                self.assert_all_called()
        finally:
            # Idempotent check, i.e. already started
            if started:
                if clear:
                    self.rollback(reset=False)
                if reset:
                    self.reset()
                self.transports.remove(self)

            self._unpatch()

    @classmethod
    def _patch(cls) -> None:
        # Ensure we only patch once!
        if cls._patches:
            return

        # Start patching target transports
        for transport in cls.targets:
            for method, new in (("request", cls._request), ("arequest", cls._arequest)):
                try:
                    spec = f"{transport}.{method}"
                    patch = mock.patch(spec, spec=True, new_callable=new)
                    patch.start()
                    cls._patches.append(patch)
                except AttributeError:
                    pass

    @classmethod
    def _unpatch(cls) -> None:
        # Ensure we don't stop patching when registered transports exists
        if cls.transports:
            return

        # Stop patching HTTPX
        while cls._patches:
            patch = cls._patches.pop()
            patch.stop()

    @classmethod
    def _request(cls, spec):
        def request(self, *args, **kwargs):
            pass_through = partial(spec, self)
            kwargs["ext"] = {**kwargs.get("ext", {}), "pass_through": pass_through}
            response = None
            error = None
            for transport in cls.transports:
                try:
                    response = transport.request(*args, **kwargs)
                except AssertionError as e:
                    error = e.args[0]
                    continue
                else:
                    break
            else:
                assert response, error
            return response

        return request

    @classmethod
    def _arequest(cls, spec):
        async def request(self, *args, **kwargs):
            pass_through = partial(spec, self)
            kwargs["ext"] = {**kwargs.get("ext", {}), "pass_through": pass_through}
            response = None
            error = None
            for transport in cls.transports:
                try:
                    response = await transport.arequest(*args, **kwargs)
                except AssertionError as e:
                    error = e.args[0]
                    continue
                else:
                    break
            else:
                assert response, error
            return response

        return request



import inspect
from collections.abc import Iterator
from typing import Any, Callable, Dict, NamedTuple, Optional, Tuple, Type, Union, cast
from unittest import mock
from warnings import warn

import httpx

from .patterns import M, Pattern
from .types import (
    ByteStream,
    ContentDataTypes,
    HeaderTypes,
    JSONTypes,
    Kwargs,
    QueryParamTypes,
    RequestTypes,
    Response,
    SideEffectTypes,
    URLPatternTypes,
)


def decode_request(request: RequestTypes) -> httpx.Request:
    """
    Build a httpx Request from httpcore request args.
    """
    if isinstance(request, httpx.Request):
        return request
    method, url, headers, stream = request
    return httpx.Request(method, url, headers=headers, stream=stream)


def encode_response(response: httpx.Response) -> Response:
    """
    Builds a raw response tuple from httpx Response.
    """
    return (
        response.status_code,
        response.headers.raw,
        response.stream,
        response.ext,
    )


def clone_response(response: httpx.Response, request: httpx.Request) -> httpx.Response:
    """
    Clones a httpx Response for given request.
    """
    response = httpx.Response(
        response.status_code,
        headers=response.headers,
        stream=response.stream,
        request=request,
        ext=dict(response.ext),
    )
    response.read()
    return response


class Call(NamedTuple):
    request: httpx.Request
    response: Optional[httpx.Response]


class CallList(list, mock.NonCallableMock):
    @property
    def called(self) -> bool:  # type: ignore
        return bool(self)

    @property
    def call_count(self) -> int:  # type: ignore
        return len(self)

    @property
    def last(self) -> Optional[Call]:
        return self[-1] if self else None

    def record(
        self, request: httpx.Request, response: Optional[httpx.Response]
    ) -> Call:
        call = Call(request=request, response=response)
        self.append(call)
        return call


class MockResponse:
    _content: Optional[ContentDataTypes]
    _text: Optional[str]
    _html: Optional[str]
    _json: Optional[JSONTypes]

    def __init__(
        self,
        status_code: Optional[int] = None,
        *,
        content: Optional[ContentDataTypes] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        headers: Optional[HeaderTypes] = None,
        content_type: Optional[str] = None,
        http_version: Optional[str] = None,
        context: Optional[Kwargs] = None,
    ) -> None:
        self.http_version = http_version
        self.status_code = status_code or 200
        self.context = context if context is not None else {}

        self.headers = httpx.Headers(headers) if headers else httpx.Headers()
        if content_type:
            self.headers["Content-Type"] = content_type

        # Set body variants in reverse priority order
        self.json = json
        self.html = html
        self.text = text
        self.content = content

    def clone(self, **context: Any) -> "MockResponse":
        merged_context = dict(self.context)
        merged_context.update(context)
        return MockResponse(
            self.status_code,
            content=self.content,
            text=self.text,
            html=self.html,
            json=self.json,
            headers=self.headers,
            http_version=self.http_version,
            context=merged_context,
        )

    def prepare(
        self,
        content: Optional[ContentDataTypes],
        *,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
    ) -> Tuple[
        Optional[ContentDataTypes], Optional[str], Optional[str], Optional[JSONTypes]
    ]:
        if content is not None:
            text = None
            html = None
            json = None
            if isinstance(content, str):
                text = content
                content = None
            elif isinstance(content, (list, dict)):
                json = content
                content = None
        elif text is not None:
            html = None
            json = None
        elif html is not None:
            json = None

        return content, text, html, json

    @property
    def content(self) -> Optional[ContentDataTypes]:
        return self._content

    @content.setter
    def content(self, content: Optional[ContentDataTypes]) -> None:
        self._content, self.text, self.html, self.json = self.prepare(
            content, text=self.text, html=self.html, json=self.json
        )

    @property
    def text(self) -> Optional[str]:
        return self._text

    @text.setter
    def text(self, text: Optional[str]) -> None:
        self._text = text
        if text is not None:
            self._content = None
            self._html = None
            self._json = None

    @property
    def html(self) -> Optional[str]:
        return self._html

    @html.setter
    def html(self, html: Optional[str]) -> None:
        self._html = html
        if html is not None:
            self._content = None
            self._text = None
            self._json = None

    @property
    def json(self) -> Optional[JSONTypes]:
        return self._json

    @json.setter
    def json(self, json: Optional[JSONTypes]) -> None:
        self._json = json
        if json is not None:
            self._content = None
            self._text = None
            self._html = None

    def as_response(self) -> httpx.Response:
        content = self._content
        context = dict(self.context)
        request = decode_request(context.pop("request"))

        if callable(content):
            if inspect.iscoroutinefunction(self._content):
                raise NotImplementedError("Async content callback no longer supported.")
            content = content(request, **context)

        if isinstance(content, Exception):
            raise content

        content, text, html, json = self.prepare(
            content, text=self.text, html=self.html, json=self.json
        )

        # Comply with httpx Response content type hints
        assert content is None or isinstance(content, bytes)

        response = httpx.Response(
            self.status_code,
            headers=self.headers,
            content=content,
            text=text,
            html=html,
            json=json,
            request=request,
        )

        if self.http_version:
            response.ext["http_version"] = self.http_version

        return response


class ResponseTemplate(MockResponse):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        warn(
            "ResponseTemplate is deprecated. Please use MockResponse.",
            category=DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class Route:
    def __init__(
        self,
        *patterns: Pattern,
        **lookups: Any,
    ) -> None:
        self.pattern = M(*patterns, **lookups)
        self.name: Optional[str] = None  # TODO: Drop or add setter to prevent change
        self.calls = CallList()
        self._return_value: Union[MockResponse, httpx.Response] = None
        self._side_effect: Optional[SideEffectTypes] = None
        self._pass_through: Optional[bool] = None
        self.snapshot()

    def __hash__(self):
        if self.pattern:
            return hash(self.pattern)
        return id(self)

    def __repr__(self):
        return f"<Route {self.pattern!r}>"  # pragma: no cover

    def __call__(self, side_effect: Callable) -> Callable:
        self.side_effect = side_effect
        return side_effect

    def __mod__(self, response: Union[int, Dict[str, Any]]) -> "Route":
        if isinstance(response, int):
            self.return_value = httpx.Response(status_code=response)

        elif isinstance(response, dict):
            response.setdefault("status_code", 200)
            self.return_value = httpx.Response(**response)

        else:
            assert isinstance(
                response, (httpx.Response, MockResponse)
            ), f"Route can only % with int or dict, got {response!r}"
            self.return_value = response

        return self

    @property
    def return_value(self) -> Optional[Union[MockResponse, httpx.Response]]:
        return self._return_value

    @return_value.setter
    def return_value(
        self, return_value: Optional[Union[MockResponse, httpx.Response]]
    ) -> None:
        self.pass_through(None)
        self._return_value = return_value

    @property
    def side_effect(self) -> Optional[SideEffectTypes]:
        return self._side_effect

    @side_effect.setter
    def side_effect(self, side_effect: Optional[SideEffectTypes]) -> None:
        self.pass_through(None)
        if not side_effect:
            self._side_effect = None
        elif isinstance(side_effect, (tuple, list, Iterator)):
            self._side_effect = iter(side_effect)
        else:
            self._side_effect = side_effect

    def snapshot(self) -> None:
        self.__return_value = self._return_value
        self.__side_effect = self._side_effect
        self.__pass_through = self._pass_through
        self._calls = CallList(self.calls)

    def rollback(self, reset: bool = True) -> None:
        self._return_value = self.__return_value
        self._side_effect = self.__side_effect
        self._pass_through = self.__pass_through
        if reset:
            self.reset()

    def reset(self) -> None:
        self.calls[:] = self._calls

    def mock(
        self,
        return_value: Optional[Union[MockResponse, httpx.Response]] = None,
        *,
        side_effect: Optional[SideEffectTypes] = None,
    ) -> "Route":
        self.return_value = return_value
        self.side_effect = side_effect
        return self

    def respond(
        self,
        status_code: int = 200,
        *,
        headers: Optional[HeaderTypes] = None,
        content: Optional[Union[str, bytes, ByteStream]] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
        json: Optional[JSONTypes] = None,
        stream: Optional[ByteStream] = None,
        **kwargs: Any,
    ) -> "Route":
        response = httpx.Response(
            status_code,
            headers=headers,
            content=content,
            text=text,
            html=html,
            json=json,
            stream=stream,
            **kwargs,
        )
        return self.mock(return_value=response)

    def pass_through(self, value: bool = True) -> "Route":
        self._pass_through = value
        return self

    @property
    def is_pass_through(self) -> bool:
        return bool(self._pass_through)

    @property
    def alias(self) -> Optional[str]:
        warn(
            ".alias property is deprecated. Please, use .name",
            category=DeprecationWarning,
        )
        return self.name

    @property
    def called(self) -> bool:
        return self.calls.called

    @property
    def call_count(self) -> int:
        return self.calls.call_count

    @property
    def stats(self):
        warn(
            ".stats property is deprecated. Please, use .calls",
            category=DeprecationWarning,
        )
        return self.calls

    def _next_side_effect(
        self,
    ) -> Union[Callable, Exception, Type[Exception], httpx.Response]:
        effect: Union[Callable, Exception, Type[Exception], httpx.Response]
        if isinstance(self._side_effect, Iterator):
            effect = next(self._side_effect)
        else:
            effect = cast(
                Union[Callable, Exception, Type[Exception]], self._side_effect
            )

        return effect

    def _call_side_effect(
        self, effect: Callable, request: httpx.Request, **kwargs: Any
    ) -> Optional[Union[httpx.Request, httpx.Response, MockResponse]]:
        argspec = inspect.getfullargspec(effect)
        if "response" in argspec.args or len(argspec.args) > 1 + len(kwargs):
            warn(
                "Side effect (callback) `response` arg is deprecated. "
                "Please instantiate httpx.Response inside your function.",
                category=DeprecationWarning,
            )
            args = (
                request,
                MockResponse(context={"request": request, **kwargs}),
            )
        else:
            args = (request,)  # type: ignore

        # Call side effect
        try:
            result = effect(*args, **kwargs)
        except Exception as error:
            raise SideEffectError(self, origin=error) from error

        # Validate result
        if result and not isinstance(
            result, (httpx.Response, MockResponse, httpx.Request)
        ):
            raise ValueError(
                f"Side effects must return; either `httpx.Response` or "
                f"`MockResponse`, `httpx.Request` for pass-through, "
                f"or `None` for a non-match. Got {result!r}"
            )

        return result

    def _resolve_side_effect(
        self, request: httpx.Request, **kwargs: Any
    ) -> Optional[Union[httpx.Request, httpx.Response, MockResponse]]:
        effect = self._next_side_effect()

        # Handle Exception `instance` side effect
        if isinstance(effect, Exception):
            raise SideEffectError(self, origin=effect)

        # Handle Exception `type` side effect
        Error: Type[Exception] = cast(Type[Exception], effect)
        if isinstance(effect, type) and issubclass(Error, Exception):
            raise SideEffectError(
                self,
                origin=(
                    Error("Mock Error", request=request)
                    if issubclass(Error, httpx.HTTPError)
                    else Error()
                ),
            )

        # Handle `Callable` side effect
        if callable(effect):
            result = self._call_side_effect(effect, request, **kwargs)
            return result

        return effect

    def resolve(
        self, request: httpx.Request, **kwargs: Any
    ) -> Optional[Union[httpx.Request, httpx.Response]]:
        result: Optional[Union[MockResponse, httpx.Response, httpx.Request]] = None

        if self._side_effect:
            result = self._resolve_side_effect(request, **kwargs)
            if result is None:
                return None  # Side effect resolved as a non-matching route

        elif self._return_value:
            result = self._return_value

        if isinstance(result, MockResponse):
            # Resolve MockResponse into httpx.Response
            try:
                result = result.clone(request=request, **kwargs)
                result = result.as_response()
            except Exception as error:
                raise SideEffectError(self, origin=error) from error

        if result is None:
            # Auto mock a new response
            result = httpx.Response(200, request=request)

        elif isinstance(result, httpx.Response) and not result._request:
            # Clone existing Response for immutability
            result = clone_response(result, request)

        return result

    def match(
        self, request: httpx.Request
    ) -> Optional[Union[httpx.Request, httpx.Response]]:
        """
        Matches and resolves request with given patterns and optional side effect.

        Returns None for a non-matching route, mocked response for a match,
        or input request for pass-through.
        """
        context = {}

        if self.pattern:
            match = self.pattern.match(request)
            if not match:
                return None
            context = match.context

        if self._pass_through:
            return request

        result = self.resolve(request, **context)
        return result


class RequestPattern(Route):
    def __init__(
        self,
        method: Union[str, Callable] = None,
        url: Optional[URLPatternTypes] = None,
        *,
        params: Optional[QueryParamTypes] = None,
        response: Optional[MockResponse] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        base_url: Optional[str] = None,
    ) -> None:
        warn(
            "RequestPattern is deprecated. Please use Route.",
            category=DeprecationWarning,
        )

        super().__init__(
            method=method if not callable(method) else None,
            base_url=base_url,
            url=url,
            params=params,
        )

        self.name = alias
        self._pass_through = pass_through

        if callable(method):
            self.side_effect = method

        if response:
            self.return_value = response


class SideEffectError(Exception):
    def __init__(self, route: Route, origin: Exception) -> None:
        self.route = route
        self.origin = origin



from typing import Any, Callable, Optional, Union, overload

from .mocks import MockTransport
from .models import CallList, Route
from .patterns import Pattern
from .types import (
    ContentDataTypes,
    DefaultType,
    HeaderTypes,
    JSONTypes,
    QueryParamTypes,
    URLPatternTypes,
)

mock = MockTransport(assert_all_called=False)

aliases = mock.routes
routes = mock.routes
stats = mock.calls
calls: CallList = mock.calls


def start() -> None:
    global mock
    mock.start()


def stop(clear: bool = True, reset: bool = True) -> None:
    global mock
    mock.stop(clear=clear, reset=reset)


def clear() -> None:
    global mock
    mock.clear()


def reset() -> None:
    global mock
    mock.reset()


@overload
def pop(name: str) -> Route:
    ...  # pragma: nocover


@overload
def pop(name: str, default: DefaultType) -> Union[Route, DefaultType]:
    ...  # pragma: nocover


def pop(name, default=...):
    global mock
    return mock.pop(name, default=default)


def route(*patterns: Pattern, name: Optional[str] = None, **lookups: Any) -> Route:
    global mock
    return mock.route(*patterns, name=name, **lookups)


def add(
    method: Union[str, Callable, Route],
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.add(
        method,
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )


def get(
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.get(
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )


def post(
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.post(
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )


def put(
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.put(
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )


def patch(
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.patch(
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )


def delete(
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.delete(
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )


def head(
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.head(
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )


def options(
    url: Optional[URLPatternTypes] = None,
    *,
    params: Optional[QueryParamTypes] = None,
    status_code: Optional[int] = None,
    headers: Optional[HeaderTypes] = None,
    content_type: Optional[str] = None,
    content: Optional[ContentDataTypes] = None,
    text: Optional[str] = None,
    html: Optional[str] = None,
    json: Optional[JSONTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
    name: Optional[str] = None,
    **lookups: Any,
) -> Route:
    global mock
    return mock.options(
        url=url,
        params=params,
        status_code=status_code,
        headers=headers,
        content_type=content_type,
        content=content,
        text=text,
        html=html,
        json=json,
        pass_through=pass_through,
        alias=alias,
        name=name,
        **lookups,
    )



import operator
import re
from enum import Enum
from functools import reduce
from http.cookies import SimpleCookie
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Pattern as RegexPattern,
    Sequence,
    Set,
    Tuple,
    Union,
)
from urllib.parse import urljoin

import httpx

from .types import (
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestTypes,
    URLPatternTypes,
)


class Lookup(Enum):
    EQUAL = "eq"
    REGEX = "regex"
    STARTS_WITH = "startswith"
    CONTAINS = "contains"
    IN = "in"


class Match:
    def __init__(self, matches: bool, **context: Any) -> None:
        self.matches = matches
        self.context = context

    def __bool__(self):
        return bool(self.matches)

    def __invert__(self):
        self.matches = not self.matches
        return self

    def __repr__(self):  # pragma: nocover
        return f"<Match {self.matches}>"


class Pattern:
    lookups: Tuple[Lookup, ...] = (Lookup.EQUAL,)
    lookup: Lookup
    key: str
    base: Optional["Pattern"]
    value: Any

    def __init__(self, value: Any, lookup: Optional[Lookup] = None) -> None:
        if lookup and lookup not in self.lookups:
            raise NotImplementedError(
                f"{lookup.value!r} is not a valid Lookup for {self.__class__.__name__!r}"
            )
        self.lookup = lookup or self.lookups[0]
        self.base = None
        self.value = self.clean(value)

    def __iter__(self):
        yield self

    def __and__(self, other: "Pattern") -> "Pattern":
        return _And((self, other))

    def __or__(self, other: "Pattern") -> "Pattern":
        return _Or((self, other))

    def __invert__(self):
        return _Invert(self)

    def __repr__(self):  # pragma: nocover
        return f"<{self.__class__.__name__} {self.lookup.value} {repr(self.value)}>"

    def __hash__(self):
        return hash((self.__class__, self.lookup, self.value))

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other)

    def clean(self, value: Any) -> Any:
        """
        Clean and return pattern value.
        """
        return value

    def parse(self, request: RequestTypes) -> Any:  # pragma: nocover
        """
        Parse and return request value to match with pattern value.
        """
        raise NotImplementedError()

    def strip_base(self, value: Any) -> Any:  # pragma: nocover
        return value

    def match(self, request: RequestTypes) -> Match:
        value = self.parse(request)

        # Match and strip base
        if self.base:
            base_match = self.base._match(value)
            if not base_match:
                return base_match
            value = self.strip_base(value)

        return self._match(value)

    def _match(self, value: Any) -> Match:
        lookup_method = getattr(self, f"_{self.lookup.value}")
        return lookup_method(value)

    def _eq(self, value: Any) -> Match:
        return Match(value == self.value)

    def _regex(self, value: str) -> Match:
        match = self.value.search(value)
        if match is None:
            return Match(False)
        return Match(True, **match.groupdict())

    def _startswith(self, value: str) -> Match:
        return Match(value.startswith(self.value))

    def _contains(self, value: Any) -> Match:  # pragma: nocover
        raise NotImplementedError()

    def _in(self, value: Any) -> Match:
        return Match(value in self.value)


class _And(Pattern):
    value: Tuple[Pattern, Pattern]

    def __repr__(self):  # pragma: nocover
        a, b = self.value
        return f"{repr(a)} AND {repr(b)}"

    def __iter__(self):
        a, b = self.value
        yield from a
        yield from b

    def match(self, request: RequestTypes) -> Match:
        a, b = self.value
        match1 = a.match(request)
        match2 = b.match(request)
        if match1 and match2:
            return Match(True, **{**match1.context, **match2.context})
        return Match(False)


class _Or(Pattern):
    value: Tuple[Pattern, Pattern]

    def __repr__(self):  # pragma: nocover
        a, b = self.value
        return f"{repr(a)} OR {repr(b)}"

    def __iter__(self):
        a, b = self.value
        yield from a
        yield from b

    def match(self, request: RequestTypes) -> Match:
        a, b = self.value
        match = a.match(request)
        if not match:
            match = b.match(request)
        return match


class _Invert(Pattern):
    value: Pattern

    def __repr__(self):  # pragma: nocover
        return f"NOT {repr(self.value)}"

    def __iter__(self):
        yield from self.value

    def match(self, request: RequestTypes) -> Match:
        return ~self.value.match(request)


class Method(Pattern):
    lookups = (Lookup.EQUAL, Lookup.IN)
    key = "method"
    value: Union[str, Sequence[str]]

    def clean(self, value: Union[str, Sequence[str]]) -> Union[str, Sequence[str]]:
        if isinstance(value, str):
            return value.upper()
        return value

    def parse(self, request: RequestTypes) -> str:
        if isinstance(request, httpx.Request):
            method = request.method
        else:
            _method, *_ = request
            method = _method.decode("ascii")
        return method


class MultiItemsMixin:
    lookup: Lookup
    value: Any

    def __hash__(self):
        return hash((self.__class__, self.lookup, str(self.value)))

    def _contains(self, value: Any) -> Match:
        value_list = self.value.multi_items()
        request_list = value.multi_items()

        if len(value_list) > len(request_list):
            return Match(False)

        for item in value_list:
            if item not in request_list:
                return Match(False)

        return Match(True)


class Headers(MultiItemsMixin, Pattern):
    lookups = (Lookup.CONTAINS, Lookup.EQUAL)
    key = "headers"
    value: httpx.Headers

    def clean(self, value: HeaderTypes) -> httpx.Headers:
        return httpx.Headers(value)

    def parse(self, request: RequestTypes) -> httpx.Headers:
        if isinstance(request, httpx.Request):
            headers = request.headers
        else:
            _, _, _headers, *_ = request
            headers = httpx.Headers(_headers)

        return headers


class Cookies(Pattern):
    lookups = (Lookup.CONTAINS, Lookup.EQUAL)
    key = "cookies"
    value: Set[Tuple[str, str]]

    def __hash__(self):
        return hash((self.__class__, self.lookup, tuple(sorted(self.value))))

    def clean(self, value: CookieTypes) -> Set[Tuple[str, str]]:
        if isinstance(value, dict):
            return set(value.items())

        return set(value)

    def parse(self, request: RequestTypes) -> Set[Tuple[str, str]]:
        if isinstance(request, httpx.Request):
            headers = request.headers
        else:
            _, _, _headers, *_ = request
            headers = httpx.Headers(_headers)

        cookie_header = headers.get("cookie")
        if not cookie_header:
            return set()

        cookies: SimpleCookie = SimpleCookie()
        cookies.load(rawdata=cookie_header)

        return {(cookie.key, cookie.value) for cookie in cookies.values()}

    def _contains(self, value: Set[Tuple[str, str]]) -> Match:
        return Match(bool(self.value & value))


class Scheme(Pattern):
    lookups = (Lookup.EQUAL, Lookup.IN)
    key = "scheme"
    value: Union[str, Sequence[str]]

    def clean(self, value: Union[str, Sequence[str]]) -> Union[str, Sequence[str]]:
        if isinstance(value, str):
            return value.lower()
        return value

    def parse(self, request: RequestTypes) -> str:
        if isinstance(request, httpx.Request):
            scheme = request.url.scheme
        else:
            _, (_scheme, *_), *_ = request
            scheme = _scheme.decode("ascii")
        return scheme


class Host(Pattern):
    lookups = (Lookup.EQUAL, Lookup.IN)
    key = "host"
    value: Union[str, Sequence[str]]

    def parse(self, request: RequestTypes) -> str:
        if isinstance(request, httpx.Request):
            host = request.url.host
        else:
            _, (_, _host, *_), *_ = request
            host = _host.decode("ascii")
        return host


class Port(Pattern):
    lookups = (Lookup.EQUAL, Lookup.IN)
    key = "port"
    value: Optional[int]

    def parse(self, request: RequestTypes) -> Optional[int]:
        scheme: Optional[str] = None
        if isinstance(request, httpx.Request):
            scheme = request.url.scheme
            port = request.url.port
        else:
            _, (_scheme, _, port, _), *_ = request
            if _scheme:
                scheme = _scheme.decode("ascii")
        scheme_port = get_scheme_port(scheme)
        return port or scheme_port


class Path(Pattern):
    lookups = (Lookup.EQUAL, Lookup.REGEX, Lookup.STARTS_WITH, Lookup.IN)
    key = "path"
    value: Union[str, Sequence[str], RegexPattern[str]]

    def clean(
        self, value: Union[str, RegexPattern[str]]
    ) -> Union[str, RegexPattern[str]]:
        if self.lookup in (Lookup.EQUAL, Lookup.STARTS_WITH) and isinstance(value, str):
            path = urljoin("/", value)  # Ensure leading slash
            value = httpx.URL(path).path
        elif self.lookup is Lookup.REGEX and isinstance(value, str):
            value = re.compile(value)
        return value

    def parse(self, request: RequestTypes) -> str:
        if isinstance(request, httpx.Request):
            path = request.url.path
        else:
            _, (_, _, _, _path), *_ = request
            _path, _, _ = _path.partition(b"?")
            path = _path.decode("ascii")
        return path

    def strip_base(self, value: str) -> str:
        value = urljoin("/", value[len(self.base.value) :])
        return value


class Params(MultiItemsMixin, Pattern):
    lookups = (Lookup.CONTAINS, Lookup.EQUAL)
    key = "params"
    value: httpx.QueryParams

    def clean(self, value: QueryParamTypes) -> httpx.QueryParams:
        return httpx.QueryParams(value)

    def parse(self, request: RequestTypes) -> httpx.QueryParams:
        if isinstance(request, httpx.Request):
            query = request.url.query
        else:
            _, url, *_ = request
            query = httpx.URL(url).query

        return httpx.QueryParams(query)  # TODO: Cache params on request?


class URL(Pattern):
    lookups = (
        Lookup.EQUAL,
        Lookup.REGEX,
        Lookup.STARTS_WITH,
    )
    key = "url"
    value: Union[str, RegexPattern[str]]

    def clean(self, value: URLPatternTypes) -> Union[str, RegexPattern[str]]:
        url: Union[str, RegexPattern[str]]
        if self.lookup is Lookup.EQUAL and isinstance(value, (str, tuple, httpx.URL)):
            _url = httpx.URL(value)
            url = str(_url)
            if not _url._uri_reference.path:  # Ensure path
                url += "/"
        elif self.lookup is Lookup.REGEX and isinstance(value, str):
            url = re.compile(value)
        elif isinstance(value, (str, RegexPattern)):
            url = value
        else:
            raise ValueError(f"Invalid url: {value!r}")
        return url

    def parse(self, request: RequestTypes) -> str:
        if isinstance(request, httpx.Request):
            url = str(request.url)
            if not request.url._uri_reference.path:  # Ensure path
                url += "/"
        else:
            _, _url, *_ = request
            url = str(httpx.URL(_url))
        return url


# TODO: Refactor this to register when subclassing
PATTERNS = {
    P.key: P
    for P in (
        Method,
        Headers,
        Cookies,
        Scheme,
        Host,
        Port,
        Path,
        Params,
        URL,
    )
}


def M(*patterns: Pattern, **lookups: Any) -> Pattern:
    extras = None

    for pattern__lookup, value in lookups.items():
        if not value:
            continue

        if pattern__lookup == "url":
            extras = parse_url_patterns(value)
            continue

        pattern_key, __, lookup_value = pattern__lookup.partition("__")
        if pattern_key not in PATTERNS:
            raise KeyError(f"{pattern_key!r} is not a valid Pattern")

        lookup = None if not lookup_value else Lookup(lookup_value)
        pattern = PATTERNS[pattern_key](value, lookup=lookup)
        patterns += (pattern,)

    pattern = combine(patterns)
    if extras:
        pattern = merge_patterns(pattern, **extras)
    return pattern


def get_scheme_port(scheme: Optional[str]) -> Optional[int]:
    return {"http": 80, "https": 443}.get(scheme)


def combine(
    patterns: Sequence[Pattern], op: Callable = operator.and_
) -> Optional[Pattern]:
    patterns = tuple(filter(None, patterns))
    if not patterns:
        return None
    return reduce(op, patterns)


def parse_url_patterns(
    url: Optional[Union[str, httpx.URL]], exact: bool = True
) -> Dict[str, Pattern]:
    bases: Dict[str, Pattern] = {}
    if not url:
        return bases

    url = httpx.URL(url)
    scheme_port = get_scheme_port(url.scheme)

    if url.scheme:
        bases[Scheme.key] = Scheme(url.scheme)
    if url.host:
        bases[Host.key] = Host(url.host)
    if url.port and url.port != scheme_port:
        bases[Port.key] = Port(url.port)
    if url._uri_reference.path:  # URL.path always returns "/"
        lookup = Lookup.EQUAL if exact else Lookup.STARTS_WITH
        bases[Path.key] = Path(url.path, lookup=lookup)
    if url.query:
        lookup = Lookup.EQUAL if exact else Lookup.CONTAINS
        bases[Params.key] = Params(url.query, lookup=lookup)

    return bases


def merge_patterns(pattern: Pattern, **bases: Pattern) -> Pattern:
    if not bases:
        return pattern

    if pattern:
        # Flatten pattern
        patterns = list(iter(pattern))

        if "host" in (_pattern.key for _pattern in patterns):
            # Pattern is "absolute", skip merging
            bases = None
        else:
            # Traverse pattern and set releated base
            for _pattern in patterns:
                base = bases.pop(_pattern.key, None)
                if base and base.lookup is Lookup.EQUAL:
                    # Skip "exact" base, pattern lookup overrides
                    continue
                _pattern.base = base

    if bases:
        # Combine left over base patterns with pattern
        base_pattern = combine(list(bases.values()))
        if pattern:
            pattern = base_pattern & pattern
        else:
            pattern = base_pattern

    return pattern



from warnings import warn

from httpcore import (
    AsyncByteStream,
    AsyncHTTPTransport,
    SyncByteStream,
    SyncHTTPTransport,
)

from .models import decode_request, encode_response
from .router import Router
from .types import URL, AsyncResponse, Headers, SyncResponse


class BaseMockTransport(Router):
    def request(
        self,
        method: bytes,
        url: URL,
        headers: Headers = None,
        stream: SyncByteStream = None,
        ext: dict = None,
    ) -> SyncResponse:
        raw_request = (method, url, headers, stream)
        request = decode_request(raw_request)
        response = self.resolve(request)

        if response is None:
            pass_through = ext.pop("pass_through", None)
            if pass_through is None:
                raise ValueError("pass_through not supported with manual transport")
            raw_response = pass_through(method, url, headers, stream, ext)
        else:
            raw_response = encode_response(response)

        return raw_response

    async def arequest(
        self,
        method: bytes,
        url: URL,
        headers: Headers = None,
        stream: AsyncByteStream = None,
        ext: dict = None,
    ) -> AsyncResponse:
        raw_request = (method, url, headers, stream)
        request = decode_request(raw_request)
        response = self.resolve(request)

        if response is None:
            pass_through = ext.pop("pass_through", None)
            if pass_through is None:
                raise ValueError("pass_through not supported with manual transport")
            raw_response = await pass_through(method, url, headers, stream, ext)
        else:
            raw_response = encode_response(response)

        return raw_response

    def close(self) -> None:
        if self._assert_all_called:
            self.assert_all_called()

    async def aclose(self) -> None:
        self.close()


class SyncMockTransport(BaseMockTransport, SyncHTTPTransport):
    def __init__(self, **kwargs):
        warn(
            "SyncMockTransport is deprecated. Please, use MockTransport",
            category=DeprecationWarning,
        )
        super().__init__(**kwargs)


class AsyncMockTransport(BaseMockTransport, AsyncHTTPTransport):
    def __init__(self, **kwargs):
        warn(
            "SyncMockTransport is deprecated. Please, use MockTransport",
            category=DeprecationWarning,
        )
        super().__init__(**kwargs)



from .__version__ import __version__
from .mocks import MockTransport
from .models import MockResponse, Route
from .router import Router
from .transports import AsyncMockTransport, SyncMockTransport

from .api import (  # isort:skip
    mock,
    routes,
    aliases,
    stats,
    calls,
    start,
    stop,
    clear,
    reset,
    pop,
    route,
    add,
    get,
    post,
    put,
    patch,
    delete,
    head,
    options,
)

__all__ = [
    "__version__",
    "MockTransport",
    "AsyncMockTransport",
    "SyncMockTransport",
    "Router",
    "Route",
    "MockResponse",
    "mock",
    "routes",
    "aliases",
    "stats",
    "calls",
    "start",
    "stop",
    "clear",
    "reset",
    "pop",
    "route",
    "add",
    "get",
    "post",
    "put",
    "patch",
    "delete",
    "head",
    "options",
]



#!/usr/bin/env python
from pathlib import Path

from setuptools import setup

exec(Path("respx", "__version__.py").read_text())  # Load __version__ into locals

setup(
    name="respx",
    version=locals()["__version__"],
    license="BSD",
    author="Jonas Lundberg",
    author_email="jonas@5monkeys.se",
    url="https://lundberg.github.io/respx/",
    keywords=["httpx", "httpcore", "mock", "responses", "requests", "async", "http"],
    description="A utility for mocking out the Python HTTPX and HTTP Core libraries.",
    long_description=Path("README.md").read_text("utf-8"),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    project_urls={
        "GitHub": "https://github.com/lundberg/respx",
        "Changelog": "https://github.com/lundberg/respx/blob/master/CHANGELOG.md",
        "Issues": "https://github.com/lundberg/respx/issues",
    },
    packages=["respx"],
    package_data={"respx": ["py.typed"]},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    install_requires=["httpx>=0.15"],
)



try:
    import pytest
except ImportError:  # pragma: nocover
    pass
else:
    import asyncio

    @pytest.fixture(scope="session")
    def session_event_loop():
        loop = asyncio.get_event_loop_policy().new_event_loop()
        yield loop
        loop.close()
